#!/usr/bin/env bash

# This script contains wrappers for Linux commands.
# Usage: $ source ./gpudb_bash_utils.sh

# Setup environment
unset COLUMNS   # truncates 'ps aux', use 'ps auxww', note PROMPT_COMMAND can reset
unset LS_COLORS # Sometimes very long

# Print debug information about status and output when running is 1.
GPUDB_DEBUG_SCRIPT=0

# ---------------------------------------------------------------------------
# Functions to initialize vars.

# All non-sourced scripts should call this with a < 16 char string
# prefaced with 'gpudb', e.g. "gpudb_COMPONENT"
# Set the '/proc/pid/comm' to identify this script, default is 'bash'.
THIS_SCRIPT_COMM=bash # global
function gpudb_set_proc_comm
{
    THIS_SCRIPT_COMM="$1"

    if ! printf "%s" "$THIS_SCRIPT_COMM" > "/proc/$$/comm" ; then
        echo_status "Script" "Error setting /proc/$$/comm to '$THIS_SCRIPT_COMM'"
        return 1
    fi
}

# All scripts that source this file should call this function in the beginning to set
# path information used by some functions in this file.
# Usage: gpudb_set_this_script_info "${BASH_SOURCE[0]}" "$@"
function gpudb_set_this_script_info
{
    local BASH_SOURCE_0="$1"
    shift

    # The directory of where this script was run from.
    THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE_0}" )" && pwd )" # NOTE: you probably want THIS_SCRIPT_LINK_DIR
    THIS_SCRIPT_NAME="${BASH_SOURCE_0##*/}"
    THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE_0}")    # Filepath that a symlink resolves to.
    THIS_SCRIPT_LINK_DIR=$(dirname "${THIS_SCRIPT_LINK}") # The actual path of the file.
    THIS_SCRIPT="$THIS_SCRIPT_DIR/$THIS_SCRIPT_NAME"      # Filepath where this was run from.
    THIS_SCRIPT_ARGS=("$@")
    #echo "THIS: '$THIS_SCRIPT_DIR' '$THIS_SCRIPT_NAME' '${THIS_SCRIPT_ARGS[@]}'"
}

# Our info, should be this filename, not caller even if sourced.
UTILS_THIS_SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
UTILS_THIS_SCRIPT_NAME="${BASH_SOURCE[0]##*/}"
UTILS_THIS_SCRIPT_LINK=$(readlink -f "${BASH_SOURCE[0]}")          # Filepath that a symlink resolves to.
UTILS_THIS_SCRIPT_LINK_DIR=$(dirname "$UTILS_THIS_SCRIPT_LINK")    # The actual directory the script resides in.
UTILS_THIS_SCRIPT="$UTILS_THIS_SCRIPT_DIR/$UTILS_THIS_SCRIPT_NAME" # Filepath where this was run from.

# ---------------------------------------------------------------------------
# Call this function to check if running as USER_NAME and if not attempt to
# use 'su' or 'sudo' to re-run this script as that user with the same args.
# This function should normally only be called by 'root' or USER_NAME,
# but does allow the owner of a given file/dir to run it for alternate installs.
# Example usage:
# gpudb_set_this_script_info "${BASH_SOURCE[0]}" "$@"
# if ! switch_user_to "$GPUDB_USER" "/opt/gpudb/core/bin" "VAR_A,VAR_B"; then
#    exit 1
# fi
# If get here, then running as user with env vars VAR_A and VAR_B else the script has exited.

function switch_user_to
{
    local USER_NAME="$1"
    local USER_FILE_OR_DIR="${2-}" # Optional path to file/dir who's owner user is allowed to run this this script, without switching user.
    local EXTRA_ENV_VARS="${3-}"   # Optional comma-separated list of env var names to pass to new shell, e.g. "A,B,C"

    local RET=0
    local CURRENT_USER=$(whoami)
    local CURRENT_USER_ID=$(id -u)
    GPUDB_RUNNING_AS_WARN="${GPUDB_RUNNING_AS_WARN:-0}"
    local USER_ID=
    if USER_ID=$(id -u "$USER_NAME" 2>&1); then
        # Check if we're already the correct user.
        if [[ "a$CURRENT_USER_ID" == "a$USER_ID" ]]; then
            return 0
        fi

        if [ -n "$USER_FILE_OR_DIR" ] ; then
            if ! FILE_DIR_OWNER_ID=$(stat -c '%u' "$USER_FILE_OR_DIR") ; then
                echo_status "Error" "Path '$USER_FILE_OR_DIR' does not exist for $THIS_SCRIPT, exiting."
                exit 1
            fi
        fi

        if [[ "a$FILE_DIR_OWNER_ID" == "a$USER_ID" ]]; then
            # Run this script again as the right user

            if [ "$EUID" -ne 0 ]; then
                # Non-root users may be asked for a password, this is their prompt.
                echo_status "Running as user" "$USER_NAME"
            fi

            local SU_OUTPUT=""
            # Try 'su' before 'sudo', which may not be installed or have config issues.
            if SU_OUTPUT=$(su -l "$USER_NAME" -c "whoami"); then
                local SU_CMD_ARGS=()
                if [ -n "$EXTRA_ENV_VARS" ]; then
                    local SU_EXTRA_ENV_VAR_ARR=()
                    IFS=, read -ra SU_EXTRA_ENV_VAR_ARR <<< "$EXTRA_ENV_VARS"
                    for v in "${SU_EXTRA_ENV_VAR_ARR[@]}" ; do
                        SU_CMD_ARGS+=("$v=\"${!v}\"") # VAR='foo' VAR2='bar' cmd...
                    done
                fi
                # Run this script again as the right user, using '-' for login to prompt for password.
                SU_CMD_ARGS+=("$THIS_SCRIPT" "${THIS_SCRIPT_ARGS[@]}") # preserve array for quoting
                local SU_CMD=$(printf '%q ' "${SU_CMD_ARGS[@]}")
                # Check for --session-command though -h help to avoid needing to ask for password again.
                #if su - $USER_NAME --session-command whoami > /dev/null 2>&1; then
                if su --help | grep -qF -- --session-command > /dev/null ; then
                    # Using '--session-command' so that ctrl-c in sleep is trapped correctly.
                    exec su - "$USER_NAME" --session-command "$SU_CMD" || RET=$?
                else
                    # Ubuntu's su doesn't have --session-command, but this version's -c works for us.
                    exec su - "$USER_NAME" -c "$SU_CMD" || RET=$?
                fi

                exit $RET # Shouldn't reach here from exec
            fi

            # Try sudo before giving up
            local SUDO_OUTPUT=""
            if SUDO_OUTPUT=$(sudo -u "$USER_NAME" "whoami" 2>&1); then
                local SUDO_PRESERVE_ENV=""
                if [ -n "$EXTRA_ENV_VARS" ]; then
                    SUDO_PRESERVE_ENV="--preserve-env=$EXTRA_ENV_VARS"
                fi
                local SUDO_CMD_ARGS=("$THIS_SCRIPT" "${THIS_SCRIPT_ARGS[@]}") # preserve array for quoting
                local SUDO_CMD=$(printf '%q ' "${SUDO_CMD_ARGS[@]}")
                exec sudo -u "$USER_NAME" -H $SUDO_PRESERVE_ENV sh -c "$SUDO_CMD" || RET=$?
            else
                echo_status "Error" "Unable to 'su' or 'sudo' from user: '$CURRENT_USER' to user: '$USER_NAME' in $THIS_SCRIPT, exiting."
                echo_status "Command" "su - $USER_NAME whoami"
                echo_status "  Output" "$SU_OUTPUT"
                echo_status "Command" "sudo -u $USER_NAME whoami"
                echo_status "  Output" "$SUDO_OUTPUT"
                exit 1
            fi

            exit $RET # Shouldn't reach here from exec

        elif [[ "a$FILE_DIR_OWNER_ID" == "a$CURRENT_USER_ID" ]]; then
            # For developer installs, we allow running as the owner of the gpudb dir.
            # This is harmless if the user owns all the dirs and they always run as this user.
            if [[ "a$GPUDB_RUNNING_AS_WARN" != "a1" ]]; then
                export GPUDB_RUNNING_AS_WARN=1
                echo_status "Warning" "Running GPUdb as user: '$CURRENT_USER' since they own '$USER_FILE_OR_DIR'"
            fi
            return 0
        else
            # This condition will probably lead to mismatched ownership of processes
            # and files requiring a chown of the GPUdb dir tree and persist dirs to fix.
            echo_status "Error" "This install should only be run by '$USER_NAME', disallowing user '$CURRENT_USER' in '$THIS_SCRIPT'"
            echo_status "" "Please check ownership of the directory tree of '$USER_FILE_OR_DIR', exiting."
            exit 1
        fi
    else
        if [[ "a$GPUDB_RUNNING_AS_WARN" != "a1" ]]; then
            export GPUDB_RUNNING_AS_WARN=1
            echo_status "Warning" "USER=$USER_NAME does not exist, running as '$CURRENT_USER' in '$THIS_SCRIPT'"
        fi
    fi

    return 0
}

# Reads the ulimit -H * and applies it to ulimit -S * if different. Soft limits can be lowered by systemd.
# Optional args are ulimits values to check in pairs of 'ulimit arg' 'min value', e.g. '-n' '1000' '-u' '2000'
function ki_ulimit_hard_to_soft
{
    # The args from 'ulimit -a', not all flags in help are handled on all hosts, check the value.
    # Skip -s stack and -i signals to avoid using too much memory.
    local ULIMIT_ARGS="-d -e -f -l -m -n -p -q -r -t -u -v -x"
    local HARD_LIM=
    local SOFT_LIM=
    for F in $ULIMIT_ARGS; do
        if  HARD_LIM=$(ulimit -H "$F" 2> /dev/null) && \
            SOFT_LIM=$(ulimit -S "$F" 2> /dev/null) && \
            [[ "a$HARD_LIM" != "a$SOFT_LIM" ]] && \
            [[ "$HARD_LIM" =~ ^(unlimited|[0-9]+)$ ]]; then
            #echo_status "Info" "Updating soft 'ulimit -S $F' from '$SOFT_LIM' to match hard limit '$HARD_LIM'"
            ulimit -S "$F" "$HARD_LIM"
        fi
    done
    # After setting hard->soft, check the provided minimum soft limits and warn.
    while [ $# -gt 1 ]; do
        local F="$1"
        local MIN_VAL="$2"
        shift; shift
        if ! SOFT_LIM=$(ulimit -S "$F" 2> /dev/null) || \
            { [[ "$SOFT_LIM" =~ ^([0-9]+)$ ]] && [ "$SOFT_LIM" -lt "$MIN_VAL" ]; }; then
            echo_status "Warning" "The soft 'ulimit -S $F' is '$SOFT_LIM', but should be at least '$MIN_VAL'"
        fi
    done
}

# ---------------------------------------------------------------------------
# The 'sleep' program doesn't propagate CTRL-C so this script may be terminated,
# but the function calling sleep continues...  Use 'read' as a replacement, but
# only when stdin (0) is a tty, otherwise 'read' immediately fails without sleeping.

if [ -t 0 ]; then
    function sleep
    {
        local TIMEOUT="$1"
        local RET_=0
        # '-N 0' so keypresses don't interrupt, just CTRL-C
        # '-s' eats keystrokes, but can break the terminal after CTRL-C, user must 'reset'
        read -t "$TIMEOUT" -N 1000000 && RET_=$? || RET_=$? # work with set -e

        if [ $RET_ -eq 130 ]; then
            exit 1 # Ctrl-C caught (128+2)
        fi

        return 0 # read timeout returns 142 (man page says > 128), return success
    }
fi

# ---------------------------------------------------------------------------
# Echo, log, stdout control

# Echo a padded string in the format "Component <pad> : State", e.g. "Text : Stopped"
GPUDB_ECHO_STATUS_WIDTH=16
function echo_status
{
    local COMPONENT="$1"
    local STATE="$2"
    printf "%-${GPUDB_ECHO_STATUS_WIDTH}s : %s\n" "$COMPONENT" "$STATE"
}

function echo_status_debug
{
    if [ "$GPUDB_DEBUG_SCRIPT" -eq 1 ]; then
        echo_status "$@"
    fi
}

# Runs the func passed in, parses out pids in the form "(1 2 3)" and echos only the numbers.
function status_to_pids
{
    local STATUS_FUNC="$1"
    local S=""
    local RET=0
    S=$($STATUS_FUNC) || RET=$? # parse even if error, some pids may be provided.
    local PIDS=
    if PIDS=$(echo "$S" | grep -oE "\([0-9 ]+\)" | sed 's/[()]//g') && \
        filter_docker_pids $PIDS ; then
        return 0
    fi
    return 1
}

function single_space_string
{
    if [[ "${1:-}" == -- ]]; then # read stdin
        tr '\n' ' ' | sed -E 's/^[ ]+//g; s/[ ]+$//g; s/ [ ]+/ /g'
    else
        echo "${1:-}" | tr '\n' ' ' | sed -E 's/^[ ]+//g; s/[ ]+$//g; s/ [ ]+/ /g'
    fi
}

# Echo the input INT_VALUE if not "" else DEFAULT_VALUE, return 1 if INT_VALUE is not an int.
function ki_check_int_or_default
{
    local INT_VALUE="$1"
    local DEFAULT_VALUE="$2"
    local MESSAGE="$3"       # 'for param --count'

    if [ -z "$INT_VALUE" ]; then
        echo "$DEFAULT_VALUE"
        return 0
    elif ! [[ "a$INT_VALUE" =~ ^a[0-9]+$ ]]; then
        >&2 echo_status "Error" "Expected an integer, but got '$INT_VALUE' $MESSAGE"
        return 1
    fi

    echo "$INT_VALUE"
}

function ki_numbers_to_ranges
{
    local NUMBERS="$1" # Space or linefeed separated "1 2 4 5 6 7 9" echos "1,2,4,5-6,9"

    local FIRST=
    local LAST=
    echo "$NUMBERS " | tr ' ' '\n' | { sort -n | uniq && echo; } | while IFS= read -r NUM; do
        # sort with extra echo to loop again to print last value
        if [ -z "$FIRST" ]; then
            FIRST="$NUM"
            LAST="$NUM"
            continue
        fi
        if [[ NUM -ne $((LAST + 1)) ]]; then
            if [[ FIRST -eq LAST ]]; then
                echo -n "$FIRST,"
            elif [[ LAST -eq $((FIRST + 1)) ]]; then
                echo -n "$FIRST,$LAST,"
            else
                echo -n "$FIRST-$LAST,"
            fi
            FIRST="$NUM"
            LAST="$NUM"
        else
            LAST=$((LAST+1))
        fi
        if [ -z "$NUM" ]; then break; fi # end of list and extra linefeed
    done | sed 's/,$//g'
}

# Get backtrace with: ki_bash_backtrace "$?" "${_:-}" "${0:-}" "${LINENO:-}"
function ki_bash_backtrace
{
    local LAST_EXIT_STATUS="$1" # $?
    local CMD="$2"              # $_
    local FILENAME="${3:-}"     # $0
    local LINE="${4:-}"         # $LINENO

    # Note "/foo.sh : line 14: $1: unbound variable" has empty '$_'
    if [[ "$CMD" == ERR ]] && [ "$LAST_EXIT_STATUS" -eq 0 ]; then LAST_EXIT_STATUS=1; fi # for trap ERR

    echo
    echo_status "Error" "Exception in ${FILENAME}:${LINE}, exit code: $LAST_EXIT_STATUS, backtrace below."
    [ -z "$CMD" ] || echo_status "Last command" "$(echo "$CMD" | head -n 3)"
    local i=0
    local C=
    while C=$(caller "$i"); do
        C=$(echo "$C" | sed -E 's/([0-9]+[ ]*)(.*$)/\2:\1/g') # put line number last
        echo_status "  $i" "$C"
        i=$((i+1))
    done

    exit "$LAST_EXIT_STATUS"
}

# Shows backtraces for unbound variables and unhandled non-0 exit code
# commands, ex. (cmd || true), (cmd || RET=$?) ...
# if 'set -o nounset', 'set -o errexit', 'set -o pipefail'
trap 'trap - ERR && ki_bash_backtrace "$?" "${_:-}" "${0:-}" "${LINENO:-}"' ERR

# ---------------------------------------------------------------------------
# Config file read/write

# Read a conf.ini parameter in the form "KEY SEPARATOR VALUE"
# Default uses 'gpudb_conf_parser.py' in same dir as this script which expands ${} variables.
# OPTIONS is a single string of:
#  --match-key    : Search for exact match of "KEY SEPARATOR *".
#                   Ignores sections, does not expand variables, and works on invalid files.
#  --match-expand[=gaia.] : Search like --match-key and if the value has a '${', use the
#                           optional '=namespace.' to prefix the key and search again using gpudb_conf_parser.py.
#  --warn-default : Print a warning if using the provided default value.
# Usage: get_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    local VALUE_GCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same
    local RET=0
    local KEYVAL_LINE=""

    if [ ! -f "$FILENAME" ]; then
        echo_status "Error" "Missing conf file: '$FILENAME'"
        ki_bash_backtrace 1 ""
        exit 1
    fi

    if [[ "$OPTIONS" == *--match-key* ]]; then
        KEYVAL_LINE=$(grep -E "^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*$SEPARATOR" "$FILENAME" | tail -n 1 | sed 's/#.*$//g; s/^[ \t]*//g; s/[ \t]*$//g' || true)
        [ -n "$KEYVAL_LINE" ] || RET=1
    elif [[ "$OPTIONS" == *--match-expand* ]]; then
        KEYVAL_LINE=$(grep -E "^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*$SEPARATOR" "$FILENAME" | tail -n 1 | sed 's/#.*$//g; s/^[ \t]*//g; s/[ \t]*$//g' || true)
        if [ -n "$KEYVAL_LINE" ]; then
            if echo "$KEYVAL_LINE" | grep -q '${'; then
                # Get the --match-expand=gaia. namespace and use the full conf parser.
                local CONF_NAMESPACE=$(echo "$OPTIONS," | sed 's/^.*--match-expand=\([^,]*\),.*/\1/g')
                KEYVAL_LINE=$("$UTILS_THIS_SCRIPT_DIR/gpudb_conf_parser.py" "$FILENAME" "${CONF_NAMESPACE}${KEY}") || RET=$?
            fi
        else
            RET=1
        fi
    else
        KEYVAL_LINE=$("$UTILS_THIS_SCRIPT_DIR/gpudb_conf_parser.py" "$FILENAME" "${KEY}") || RET=$?
    fi

    if [ $RET -ne 0 ] || [ -z "$KEYVAL_LINE" ]; then
        if [ $# -gt 4 ] && [[ "a$DEFAULT_VALUE" != "a$get_conf_file_property_NO_DEFAULT" ]]; then
            if [[ "$OPTIONS" == *--warn-default* ]]; then
                echo_status "Warning" "Missing param in '$FILENAME', using default '$KEY $SEPARATOR $DEFAULT_VALUE'"
            fi
            VALUE_GCFP_="$DEFAULT_VALUE"
        else
            echo_status "Error" "Missing conf param '$KEY' in file '$FILENAME' in '$THIS_SCRIPT', exiting."
            ki_bash_backtrace 1 ""
            exit 1
        fi
    else
        VALUE_GCFP_=$(echo "$KEYVAL_LINE" | cut -d "$SEPARATOR" -f2- | sed 's/^[ \t]*//g; s/[ \t]*$//g')
    fi

    eval "$OUTPUT_VAR_NAME"="'$VALUE_GCFP_'"
}
# Constants for get_conf_file_property
get_conf_file_property_OPTIONS="--match-key|--match-expand|--warn-default"
get_conf_file_property_NO_DEFAULT="X-get_conf_file_property_NO_DEFAULT"

# Read a conf.ini parameter in the form "KEY SEPARATOR VALUE" using grep.
# Does not fail on bad files, ignores the rest of the file.
# Note that this function ignores [ini] style sections, do not prefex keys with them.
# Usage: get_conf_file_property_simple [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]

function get_conf_file_property_simple
{
    get_conf_file_property "--match-key" "$@"
}

# Get a key='true/false' conf property, set OUTPUT_VAR_NAME to 1/0
# See get_conf_file_property() for OPTIONS.
# Usage: get_true_false_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_true_false_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    local RET=0
    VALUE_GTFCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GTFCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE" || RET=$?

    # Using grep since "tr [:upper:] [:lower:]" gives an error when run on some
    # systems in this script, but not as the user on the command line.
    if echo "$VALUE_GTFCFP_" | grep -qi '^true$' ; then
        VALUE_GTFCFP_=1
    elif echo "$VALUE_GTFCFP_" | grep -qi '^false$' ; then
        VALUE_GTFCFP_=0
    else
        echo_status "Error" "Invalid value in conf file: '$FILENAME' for param: '$KEY=$VALUE_GTFCFP_', expected either 'true' or 'false' in '$THIS_SCRIPT', exiting."
        ki_bash_backtrace 1 ""
        exit 1
    fi

    eval "$OUTPUT_VAR_NAME"="'$VALUE_GTFCFP_'"
    return $RET
}

# Get a key='123' integer number conf property, set OUTPUT_VAR_NAME to value.
# See get_conf_file_property() for OPTIONS.
# Usage: get_int_conf_file_property [OPTIONS] VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_int_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    local RET=0
    VALUE_GICFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GICFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE" || RET=$?

    if ! echo "$VALUE_GICFP_" | grep -qE '^[0-9]+$' ; then
        echo "Error" "Invalid value in conf file: '$FILENAME' for param: '$KEY=$VALUE_GICFP_', expected an integer in '$THIS_SCRIPT', exiting."
        ki_bash_backtrace 1 ""
        exit 1
    fi

    eval "$OUTPUT_VAR_NAME"="'$VALUE_GICFP_'"
    return $RET
}

# Get a conf property that is a directory, handle relative paths, and verify.
# Usage: get_directory_conf_file_property VARIABLE_NAME "KEY_NAME" "=" conf.ini [OPTIONAL_DEFAULT_VALUE_IF_MISSING]
function get_directory_conf_file_property
{
    local OPTIONS=
    while echo "$1" | grep -qE -- "$get_conf_file_property_OPTIONS"; do
        OPTIONS+="$1,"
        shift
    done
    OPTIONS="${OPTIONS// /}" # remove spaces

    local OUTPUT_VAR_NAME="$1"
    local KEY="$2"
    local SEPARATOR="$3"
    local FILENAME="$4"
    local DEFAULT_VALUE="${5-$get_conf_file_property_NO_DEFAULT}"

    local RET=0
    VALUE_GDCFP_="" # unique name, no result if OUTPUT_VAR_NAME is the same or local
    get_conf_file_property $OPTIONS VALUE_GDCFP_ "$KEY" "$SEPARATOR" "$FILENAME" "$DEFAULT_VALUE" || RET=$?

    PATH_GDCFP_="$VALUE_GDCFP_"
    fix_gpudb_conf_path PATH_GDCFP_ "$VALUE_GDCFP_" "$FILENAME"

    if [ ! -d "$PATH_GDCFP_" ]; then
        if [[ "a$PATH_GDCFP_" != "a$VALUE_GDCFP_" ]]; then
            echo_status "Error" "Missing or invalid conf param value for directory '$KEY=$VALUE_GDCFP_' which resolves to '$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        else
            echo_status "Error" "Missing or invalid conf param value for directory '$KEY=$PATH_GDCFP_' in '$THIS_SCRIPT', exiting."
        fi
        ki_bash_backtrace 1 ""
        exit 1
    fi

    eval "$OUTPUT_VAR_NAME"="'$PATH_GDCFP_'"
    return $RET
}

# Change a conf.ini parameter in the form "KEY = *" to "KEY = VAL" in FILENAME.
# Usage: change_conf_file_property "KEY_NAME" "=" "NEW_VALUE" conf.ini
function change_conf_file_property
{
    local KEY="$1"
    local SEPARATOR="$2"
    local NEW_VALUE="$3"
    local FILENAME="$4"

    # Also verify that it was written since sed always returns 0.
    if [[ $KEY == *"\n"* ]]; then
        eval "sed -i \"N;s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        if ! eval "grep -qEo \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null ; then
            echo_status "Error" "Unable to update conf '$FILENAME', '$KEY${SEPARATOR}$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            ki_bash_backtrace 1 ""
            exit 1
        fi
    else
        eval "sed -i \"s@\(^[ \t]*${KEY}[ \t]*${SEPARATOR}\).*@\1${NEW_VALUE}@\" $FILENAME"
        if ! eval "grep -qE \"^[ $(echo -e '\t')]*${KEY}[ $(echo -e '\t')]*${SEPARATOR}${NEW_VALUE}\" $FILENAME" > /dev/null; then
            echo_status "Error" "Unable to update conf '$FILENAME', '$KEY${SEPARATOR}$NEW_VALUE' in '$THIS_SCRIPT', exiting."
            ki_bash_backtrace 1 ""
            exit 1
        fi
    fi
}

# Turn a relative path from the conf file into an absolute path
function fix_gpudb_conf_path
{
    local OUTPUT_VAR_NAME="$1"
    local INPUT_FILE_PATH="$2"
    local CONF_FILE="$3"

    if [[ "$INPUT_FILE_PATH" != /* ]]; then
        # DO NOT call readlink in case it is a symlink and there's an error.
        # It might be more confusing that the dir has changed from the conf path.
        INPUT_FILE_PATH="$(dirname "$CONF_FILE")/$INPUT_FILE_PATH"
    fi

    eval "$OUTPUT_VAR_NAME"="'$INPUT_FILE_PATH'"
}

# ---------------------------------------------------------------------------
# File directory validation

function gpudb_check_temp_directory
{
    local GPUDB_CONF_FILE=$(readlink -m "$1")

    local GPUDB_CONF_TEMP_DIRECTORY=
    get_conf_file_property --match-expand=gaia. GPUDB_CONF_TEMP_DIRECTORY temp_directory '=' "$GPUDB_CONF_FILE"
    if [ -z "$GPUDB_CONF_TEMP_DIRECTORY" ]; then
        echo_status "Error" "The conf param 'temp_directory' is empty in $GPUDB_CONF_FILE, please specify a writable directory."
        exit 1
    fi
    # Path might be relative to conf file
    local GPUDB_CONF_TEMP_DIRECTORY_ABS="$GPUDB_CONF_TEMP_DIRECTORY"
    fix_gpudb_conf_path GPUDB_CONF_TEMP_DIRECTORY_ABS "$GPUDB_CONF_TEMP_DIRECTORY" "$GPUDB_CONF_FILE"
    if [ ! -d "$GPUDB_CONF_TEMP_DIRECTORY_ABS" ] || \
       [ ! -w "$GPUDB_CONF_TEMP_DIRECTORY_ABS" ]; then
        echo_status "Creating" "$(basename "$GPUDB_CONF_FILE") temp_directory $GPUDB_CONF_TEMP_DIRECTORY_ABS"
        mkdir -p "$GPUDB_CONF_TEMP_DIRECTORY"
        # The gpudb_proc user might use this directory too.
        chmod 0775 "$GPUDB_CONF_TEMP_DIRECTORY"
        chown gpudb:gpudb_proc "$GPUDB_CONF_TEMP_DIRECTORY"
    fi

    if [ ! -d "$GPUDB_CONF_TEMP_DIRECTORY_ABS" ] || \
       [ ! -w "$GPUDB_CONF_TEMP_DIRECTORY_ABS" ]; then
        echo_status "Error" "Missing and unable to create the temp_directory=$GPUDB_CONF_TEMP_DIRECTORY_ABS specified in $GPUDB_CONF_FILE"
        return 1
    fi
    return 0
}

function gpudb_check_dir
{
    if [ ! -d "$1" ]; then
        echo "Error: Missing directory: '$1' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

function gpudb_check_file
{
    if [ ! -f "$1" ]; then
        echo "Error: Missing file: '$1' in '$THIS_SCRIPT', exiting."
        exit 1
    fi
}

# Delete files in DIR matching FILE_WILDCARD, older than DAYS.
function ki_rm_old_files
{
    local RM_CMD="rm -f"
    if [ "$1" == --dry-run ]; then
        RM_CMD="ls -l"
        shift
    fi

    local DAYS="$1"     # Delete the files older than this many days.
    local DIR="$2"      # Path to find files in.
    local WILDCARD="$3" # Use empty string for *, else wilcard suitable for 'find' command.

    ki_check_int_or_default "$DAYS" "" "for the number of days to delete older files" > /dev/null || return 1
    if [ ! -d "$DIR" ]; then
        return 0
    fi
    local MTIME_PARAM="-mtime +$DAYS"
    local NAME_PARAM=$([ -z "$WILDCARD" ] || echo "-name $WILDCARD")

    # note: find returns 0 if no files found
    find "$DIR" -type f $NAME_PARAM $MTIME_PARAM -print0 | xargs -0 --no-run-if-empty $RM_CMD || return 1
    #find "$DIR" -mindepth 1 -type d -empty -delete # delete empty child dirs
}

# Delete oldest files in DIR matching FILE_WILDCARD if more than MAX_COUNT.
function ki_rm_old_if_too_many
{
    local RM_CMD="rm -f --"
    if [ "$1" == --dry-run ]; then
        RM_CMD="ls -l --"
        shift
    fi
    local MAX_COUNT="$1" # Delete the oldest files to get to max count total files.
    local DIR="$2"       # Path to find files in.
    local WILDCARD="$3"  # Use empty string for *, else wilcard suitable for 'find' command.

    ki_check_int_or_default "$MAX_COUNT" "" "for the max number of files" > /dev/null || return 1
    if [ ! -d "$DIR" ]; then
        return 0
    fi

    pushd "$DIR" > /dev/null || return 1
        if [ -z "$WILDCARD" ]; then
            \ls -1tp  | grep -v '/$' | tail -n "+$MAX_COUNT" | tr '\n' '\0' | xargs -0 --no-run-if-empty $RM_CMD
        else
            find . -name "$WILDCARD" -printf "%T@ %p\n" | sort -rn  | cut -d' ' -f2 | tail -n "+$MAX_COUNT" | tr '\n' '\0' | xargs -0 --no-run-if-empty $RM_CMD
        fi
    popd > /dev/null
}

# Delete oldest files in DIR matching FILE_WILDCARD if total size is more than MAX_SIZE in bytes.
function ki_rm_old_if_too_large
{
    local RM_CMD="rm -f --"
    if [ "$1" == --dry-run ]; then
        RM_CMD="ls -l --"
        shift
    fi
    local MAX_SIZE="$1"  # Delete the oldest files to get to total size less than max bytes.
    local DIR="$2"       # Path to find files in.
    local WILDCARD="$3"  # Use empty string for *, else wilcard suitable for 'find' command.

    ki_check_int_or_default "$MAX_SIZE" "" "for the max file size" > /dev/null || return 1
    if [ ! -d "$DIR" ]; then
        return 0
    fi
    local NAME_PARAM=$([ -z "$WILDCARD" ] || echo "-name $WILDCARD")

    pushd "$DIR" > /dev/null || return 1
        # Note: This size may be less than 'du' due to sparseness, %b is 512byte blocks.
        find . $NAME_PARAM -type f -printf '%T@ %b :%p\0' | sort -zrn | \
          awk -vRS='\0' -vORS='\0' '{du += 512*$2} du > '"$MAX_SIZE"' { sub("[^:]*:",""); print }' | \
          xargs -0 --no-run-if-empty $RM_CMD || return 1
    popd > /dev/null
}

# ---------------------------------------------------------------------------
# Networking

# Checks if a port is used, logs error and exits 1.
function check_port_usage
{
    local PORT="$1"
    local ERR_MSG="$2"

    if netstat -tuln | grep ":$PORT " | grep -q LISTEN; then
        echo "$ERR_MSG"
        NETSTAT_TXT=$(netstat -tulpn | grep ":$PORT " 2> /dev/null || true)
        if [[ "a$NETSTAT_TXT" != "a" ]]; then
            echo "$NETSTAT_TXT"
            ps uwwp $(echo "$NETSTAT_TXT" | awk '{print $7}' | grep -oE '[0-9]+') || true
        else
            netstat -tuln | grep "$PORT" || true
            echo "Unable to determine other app, please run 'netstat -tulpn | grep \":$PORT \"' as root to find PID"
        fi
        exit 1
    fi
}

# ---------------------------------------------------------------------------
# Pids

# Prints the given parent and child PIDs from a quoted string of parent PIDs.
function get_child_pids
{
    local PARENT_PIDS="$1" # space-separated list of pids, "1 2 3"
    [ -n "$PARENT_PIDS" ] || return 1

    local CHILD_PIDS=""
    for PID in $PARENT_PIDS; do
        local B=$(ps -ww --no-header -o pid,ppid -e | awk "-vp=$PID" 'function r(s){print s;s=a[s];while(s){sub(",","",s);t=s;sub(",.*","",t);sub("[0-9]+","",s);r(t)}}{a[$2]=a[$2]","$1}END{r(p)}')
        local C=$(ps -ww --no-header -o pid $B 2> /dev/null || true) # double check, doesn't print if missing.
        CHILD_PIDS+=" $C"
    done

    CHILD_PIDS=$(echo "${CHILD_PIDS}" | tr ' ' '\n' | sort -n | uniq)
    single_space_string "$CHILD_PIDS"
}

# Prints the given child and parent pids up to pid:0 from a quoted string of child PIDs.
function get_parent_pids
{
    local CHILD_PIDS="$1" # space-separated list of pids, "1 2 3"
    [ -n "$CHILD_PIDS" ] || return 1

    local PARENT_PIDS=""
    for PID in $CHILD_PIDS; do
        local B=$(ps -ww --no-header -o ppid,pid -e | awk "-vp=$PID" 'function r(s){print s;s=a[s];while(s){sub(",","",s);t=s;sub(",.*","",t);sub("[0-9]+","",s);r(t)}}{a[$2]=a[$2]","$1}END{r(p)}')
        local C=$(ps -ww --no-header -o pid $B 2> /dev/null || true) # double check, doesn't print if missing.
        PARENT_PIDS+=" $C"
    done

    PARENT_PIDS=$(echo "${PARENT_PIDS}" | tr ' ' '\n' | sort -n | uniq)
    single_space_string "$PARENT_PIDS"
}

# Echos PIDs that match grep -E regex patterns in 'ps auxwww' and /proc/PID/environ.
# Returns empty string and exit code 1 if no PIDs. Does not return self pid ($$) or parents.
# Ex: get_pids_for_app "/path/to/app --config /path/to/config" "KINETICA_PROCESS_ID=$KINETICA_PROCESS_ID"
function get_pids_for_app
{
    local PARENT_ONLY= # If 'parent' return only a single pid.
    if [[ "$1" == "--parent" ]]; then
        PARENT_ONLY="parent"
        shift
    fi

    local PS_GREP="$1"
    local ENV_GREP="$2"
    PARENT_ONLY="${3:-$PARENT_ONLY}" # If 'parent' return only a single pid.

    local P=
    if ! P=$(ps --no-header auxwww | grep -E "$PS_GREP" | grep -v 'grep -E' | awk '{print $2}' | grep -v "^$$$") || \
        [ -z "$P" ]; then
        return 1
    fi
    #>&2 echo "DEBUG get_pids_for_app '$PS_GREP' '$ENV_GREP' '$P'"

    local P1="$P"
    # Filter out all processes that don't match given env var
    if [ -n "$ENV_GREP" ] && [[ "$ENV_GREP" != . ]] ; then
        local P_ENV_PATHS=$(echo "$P" | awk '{printf "/proc/"$1"/environ "}')
        if ! P1=$({ grep -E --binary --null-data --with-filename "$ENV_GREP" $P_ENV_PATHS 2>/dev/null || true; } | tr '\0' '\n' | sed 's@^/proc/\([0-9]*\)/environ.*$@\1@g' | grep -oE '^[0-9]+$') || \
            [ -z "$P1" ]; then
            return 1
        fi
    fi

    if ! P1=$(filter_docker_pids $(echo "$P1" | tr '\n' ' ')); then
        return 1
    fi

    # Expand list to get children to avoid zombies.
    if [[ "$PARENT_ONLY" != "parent" ]]; then
        if ! P1=$(get_child_pids "$P1"); then
            return 1
        fi
    fi
    # Remove, at least, the extra pid from running $(get_child_pids P1) above.
    if ! P1=$(ki_update_pids "$P1"); then
        return 1
    fi

    # Remove this process $$ and parents, but allow children.
    local PARENT_PIDS_RE=$(get_parent_pids "$$" | tr ' ' '|')
    if ! P1=$(echo "$P1" | tr ' ' '\n' | grep -vE "^($PARENT_PIDS_RE)$") || [ -z "$P1" ]; then
        return 1
    fi
    #>&2 echo "DEBUG get_pids_for_app > '$P1'"
    #ps auxwwf | grep -oE "^[^ ]+[ ]+($(single_space_string "$P1"|tr ' ' '|')).*" >&2

    # ps -q 'pids' must be single spaced.
    single_space_string "$P1"
}

# Prints only currently running pids in list, returns 1 if none are running.
function ki_update_pids
{
    local PIDS=$(single_space_string "$1") # Space separated list of pids for ps -p
    [ -n "$PIDS" ] || return 1
    local RUNNING_PIDS=$(ps -p "$PIDS" --no-header -o pid | single_space_string -- || true)
    [ -n "$RUNNING_PIDS" ] || return 1
    echo "$RUNNING_PIDS"
}

# Signal term to pids, then term to pids+children, then kill -9 to all pids+children
# Options are arg 1 and comma separated.
function ki_stop_pids
{
    local TERM_SLEEP=4 # Seconds after term to wait before kill
    local KILL_SLEEP=4 # Seconds to wait after kill for pids to be gone.

    local HAS_OPTION=0
    local APP_NAME=
    local PROGRESS_CHAR=
    if [[ "$1" == *--name=* ]]; then
        APP_NAME=$(echo "$1" | grep -o -- "--name=[^,]*" | cut -d= -f2-)
        HAS_OPTION=1
    fi
    if [[ "$1" == *--progress=* ]]; then
        # String to 'echo -n' without linefeed.
        PROGRESS_CHAR=$(echo "$1" | grep -o -- "--progress=[^,]*" | cut -d= -f2-)
        HAS_OPTION=1
    fi
    if [[ "$1" == *--timeout=* ]]; then
        TERM_SLEEP=$(echo "$1" | grep -o -- "--timeout=[^,]*" | cut -d= -f2-)
        HAS_OPTION=1
    fi
    if [ $HAS_OPTION -eq 1 ]; then
        shift
    fi

    local PIDS=$(single_space_string "${1:-}")  # One or more space separated

    # -------------

    if [ -z "$PIDS" ]; then
        if [ -n "$APP_NAME" ]; then
            echo_status "$APP_NAME" "Stopped"
        fi
        return 0
    fi

    local CHILD_PIDS=$(get_child_pids "$PIDS") # check children to not orphan or zombie them
    if [ -n "$APP_NAME" ] && [ "$TERM_SLEEP" -gt 30 ]; then
        if [ -n "$PROGRESS_CHAR" ]; then
            echo -n "$(echo_status "$APP_NAME" "Stopping ($PIDS) ")"
        else
            echo_status "$APP_NAME" "Stopping ($PIDS)"
        fi
    fi

    # Send SIGTERM to each given pid first
    for p in $PIDS; do
        if ! ps -d -q "$p" &> /dev/null; then # -d excludes session leaders
            # We are a session leader, killing session kills all children.
            kill -- "-$p" &> /dev/null || true
        else
            kill "$p" &> /dev/null || true
        fi
    done
    # Wait for all children
    if [ -n "$PROGRESS_CHAR" ]; then
        ki_wait_for_cmd "$TERM_SLEEP" "ps -q '$CHILD_PIDS'" 0 || true
    else
        ki_wait_for_cmd "$TERM_SLEEP" "ps -q '$CHILD_PIDS'" 0 > /dev/null || true
    fi
    if CHILD_PIDS=$(ki_update_pids "$CHILD_PIDS"); then
        # Send one last term to all remaining children
        kill $CHILD_PIDS &> /dev/null || true
        sleep 1
    fi
    # Finally, kill -9
    if CHILD_PIDS=$(ki_update_pids "$CHILD_PIDS"); then
        kill -9 $CHILD_PIDS &> /dev/null || true
        if [ -n "$PROGRESS_CHAR" ]; then
            ki_wait_for_cmd "$KILL_SLEEP" "ps -q '$CHILD_PIDS'" 0 || true
        else
            ki_wait_for_cmd "$KILL_SLEEP" "ps -q '$CHILD_PIDS'" 0 > /dev/null || true
        fi
        if CHILD_PIDS=$(ki_update_pids "$CHILD_PIDS"); then
            kill -9 $CHILD_PIDS &> /dev/null || true
            sleep 4
        fi
    fi
    if [ -n "$PROGRESS_CHAR" ]; then
        echo
    fi

    if CHILD_PIDS=$(ki_update_pids "$CHILD_PIDS"); then
        if [ -n "$APP_NAME" ]; then
            echo_status "$APP_NAME" "Error: Unable to stop, processes are running ($CHILD_PIDS)"
        else
            echo "Error: Unable to stop, processes are running ($CHILD_PIDS)"
        fi
        return 1
    fi

    if [ -n "$APP_NAME" ]; then
        echo_status "$APP_NAME" "Stopped"
    fi
    return 0
}

# Execute one or two commands until their exit code chagnes.
function ki_wait_for_cmd
{
    local VERBOSE=0
    if [[ "$1" == --verbose ]]; then
        shift
        VERBOSE=1
    fi

    local TIMEOUT="$1"                 # Seconds to wait for CMD to return CMD_RET.
    local CMD1="$2"                    # cmd to run and wait as long as
    local CMD1_WAIT_IF_RET_EQ="$3"     # it's exit code equals this value.
    local CMD2="${4:-}"                # Optional second cmd to run and wait as long as
    local CMD2_WAIT_IF_RET_EQ="${5:-}" # it's exit code equals this value.

    local EXIT_PROMPT_TIME_SECS=5
    local counter=0

    function ki_wait_for_cmd_
    {
        local CMD_N="$1"
        local CMD="$2"
        local CMD_WAIT_IF_RET_EQ="$3"
        local CMD_TO_RUN="$CMD"

        if [ -z "$CMD" ]; then
            return
        elif [ $VERBOSE -eq 0 ]; then
            CMD_TO_RUN="{ $CMD; } &> /dev/null"
        fi
        if eval "$CMD_TO_RUN" ; then
            if [ "$CMD_WAIT_IF_RET_EQ" -ne 0 ] ; then
                echo_status_debug "wait_for_cmd $CMD_N" "$CMD complete, returned 0"
                return 1
            fi
        elif [ "$CMD_WAIT_IF_RET_EQ" -eq 0 ] ; then
            echo_status_debug "wait_for_cmd $CMD_N" "$CMD complete, returned 1"
            return 1
        fi
        echo_status_debug "wait_for_cmd $CMD_N" "$CMD"
    }

    while [ "$counter" -le "$TIMEOUT" ]; do
        ki_wait_for_cmd_ 1 "$CMD1" "$CMD1_WAIT_IF_RET_EQ" || return 0
        ki_wait_for_cmd_ 2 "$CMD2" "$CMD2_WAIT_IF_RET_EQ" || return 0

        if [ $((counter % EXIT_PROMPT_TIME_SECS)) -eq 0 ]; then
            echo -n "."
        fi

        sleep 1
        counter=$((counter+1))
    done
    return 1
}

# Kill all children of this process, ctrl-C during sleep can create orphan jobs.
# Add the 'trap' line below to your script to do process cleanup.
# trap "trap - SIGHUP SIGINT SIGTERM && kill_children" SIGHUP SIGINT SIGTERM
function kill_children
{
    local CHILD_PIDS=$(jobs -pr || true)
    local ORIGINAL_CHILD_PIDS=$(get_child_pids "$$ $CHILD_PIDS" || true)
    ORIGINAL_CHILD_PIDS=$(echo "$ORIGINAL_CHILD_PIDS" | tr ' ' '\n' | sed "s/^$$$//g" | single_space_string -- )

    for i in 1 2 3 4; do
        if ! CHILD_PIDS=$(jobs -pr) || [ -z "$CHILD_PIDS" ]; then
            break
        fi

        if [ $i -gt 3 ]; then
            echo_status "Terminating" "$CHILD_PIDS"
            jobs -prl || true
        fi
        kill $CHILD_PIDS &> /dev/null || true
        sleep 2
    done

    if ps -p "$ORIGINAL_CHILD_PIDS" &> /dev/null; then # true if any exist
        kill $ORIGINAL_CHILD_PIDS &> /dev/null || true
        sleep 2
        kill -9 $ORIGINAL_CHILD_PIDS &> /dev/null || true
    fi

    exit 0 # exit from trap
}

# Returns success if currently running in docker, else return 1.
IS_RUNNING_IN_DOCKER=  # 1 if true, else 0 after calling get_is_running_in_docker
function get_is_running_in_docker
{
    if [ -z "$IS_RUNNING_IN_DOCKER" ]; then
        IS_RUNNING_IN_DOCKER=0
        if grep -qF "/docker" "/proc/$$/cgroup" ; then
            IS_RUNNING_IN_DOCKER=1
        fi
    fi

    [ $IS_RUNNING_IN_DOCKER -eq 1 ] && return 0 || return 1
}

# Echos the input list of pids and if running on host removes pids of docker processes.
# Returns 1 if pid list is empty
function filter_docker_pids
{
    local PIDS="${*:-}"         # Space-separated list of pids
    local VALID_PIDS="$PIDS"

    if [ -n "$PIDS" ] && ! get_is_running_in_docker; then
        VALID_PIDS=""
        for pid in $PIDS; do
            if [ -f "/proc/${pid}/cgroup" ] && ! grep -qF "/docker" "/proc/${pid}/cgroup" &> /dev/null ; then
                VALID_PIDS+="$pid "
            fi
        done
    fi

    VALID_PIDS=$(single_space_string "$VALID_PIDS")
    if [ -z "$VALID_PIDS" ]; then
        return 1
    fi

    echo -n "$VALID_PIDS"
}

# ---------------------------------------------------------------------------

function ki_app_enabled_installed
{
    local APP_NAME="$1"
    local ENABLED_FUNC="$2"
    local INSTALLED_FUNC="$3"

    if ! $ENABLED_FUNC ; then
        echo_status "$APP_NAME" "Not enabled"
        return 1
    elif ! $INSTALLED_FUNC; then
        echo_status "$APP_NAME" "Not installed"
        return 1
    fi
}

function ki_app_stop
{
    local APP_NAME="$1" # For 'echo_status APP_NAME Stopped'
    local TIMEOUT="$2"  # Seconds to wait after term before kill.
    local PIDS="$3"     # Pid of parent or of all processes.

    local APP_PARENT=
    if APP_PARENT=$(ki_app_parent "$PIDS"); then
        echo_status "Warning" "$APP_NAME is managed by $APP_PARENT. Terminating, $APP_PARENT will restart."
    fi

    ki_stop_pids "--name=$APP_NAME,--progress=.,--timeout=$TIMEOUT" "$PIDS" || return 1
}

function ki_app_status
{
    local APP_NAME="$1"      # For 'echo_status APP_NAME Stopped'
    local PIDS="$2"          # Pid of parent or of all processes.
    local INSTALLED="${3:-}" # Either function name or 0/1
    local ENABLED="${4:-}"   # Either function name or 0/1

    local EXTRA_STATUS=
    if { [[ "$INSTALLED" =~ [a-z]+ ]] && ! "$INSTALLED"; } || [ "$INSTALLED" -eq 0 ] &> /dev/null; then
        EXTRA_STATUS=" (Not installed)"
    elif { [[ "$ENABLED" =~ [a-z]+ ]] && ! "$ENABLED"; } || [ "$ENABLED" -eq 0 ] &> /dev/null; then
        EXTRA_STATUS=" (Not enabled)"
    fi

    if [ -z "$PIDS" ]; then
        echo_status "$APP_NAME" "Stopped$EXTRA_STATUS"
        return 1
    fi
    echo_status "$APP_NAME" "Running ($PIDS)$EXTRA_STATUS"
}

# Returns 'host-manager' or 'supervisor' if any PID has it for a parent.
function ki_app_parent
{
    local PIDS="${1:-}"
    if [ -z "$PIDS" ]; then return 1; fi
    local PPID_=
    for PID in $PIDS; do
        while PPID_=$(ps --no-header -www -o ppid=,cmd= -q "$PID" 2>&1); do
            if [[ "$PPID_" == */bin/gpudb-host-manager* ]]; then
                echo "host-manager"
                return
            elif [[ "$PPID_" == */bin/supervisord* ]]; then
                echo "supervisor"
                return
            fi
            PID=$(echo "$PPID_" | awk '{print $1}')
        done
    done
    return 1
}

# ---------------------------------------------------------------------------
# supervisord helper functions

# Only set the variables.
function ki_supervisor_init
{
    KI_SUPERVISOR_DIR=$(readlink -m "$1") # "$GPUDB_ROOT_DIR/supervisor"
    KI_SUPERVISORCTL_EXE="$KI_SUPERVISOR_DIR/bin/supervisorctl"
    KI_SUPERVISORD_EXE="$KI_SUPERVISOR_DIR/bin/supervisord"
    KI_SUPERVISORD_CONF="$KI_SUPERVISOR_DIR/conf/supervisord.conf"
    KI_SUPERVISORD_SOCK="$KI_SUPERVISOR_DIR/var/supervisor.sock"
}

# Minimum supervisorctl cmd, prefer using ki_supervisorctl()
function ki_supervisorctl_exec
{
    "$KI_SUPERVISORCTL_EXE" -c "$KI_SUPERVISORD_CONF" "$@" || return 1
}

function ki_supervisorctl
{
    local VERBOSE=0
    if [[ "$1" == "--verbose" ]]; then
        shift
        VERBOSE=1
    fi

    if [[ "$1" == "pid" ]]; then
        shift; ki_supervisorctl_pid "$@" && return 0 || return $?
    elif [[ "$1" == "status" ]]; then
        shift; ki_supervisorctl_status "$@" && return 0 || return $?
    elif [[ "$1" == "stop" ]]; then
        shift; ki_supervisorctl_stop "$@" && return 0 || return $?
    fi

    # Start the supervisor demon if not running, only if necessary.
    if ! ki_supervisord_status > /dev/null; then
        ki_supervisord_start || exit 1
    fi

    local SUPERVISORCTL_OUT=
    if ! SUPERVISORCTL_OUT=$(ki_supervisorctl_exec "$@" 2>&1); then
        echo_status "Error" "Executing: $KI_SUPERVISORCTL_EXE $*"
        echo "$SUPERVISORCTL_OUT" | sed 's/^/    /g'
        return 1
    fi

    if [ $VERBOSE -eq 1 ]; then
        echo "$SUPERVISORCTL_OUT"
    fi
}

function ki_supervisorctl_pid
{
    ki_supervisord_status > /dev/null || return 1

    # Returns 7 if one app isn't running, parse anyway.
    local SUPERVISORCTL_OUT=$(ki_supervisorctl_exec pid "$@" || true)
    # One pid per line, remove lines '0' for not running, keep rest that are only numbers.
    local P=$(echo "$SUPERVISORCTL_OUT" | grep -E "^[0-9]+$" | sed 's/^0$//g' | single_space_string --)
    [ -n "$P" ] || return 1 # Empty list
    echo "$P" # Might be partial list
}

function ki_supervisorctl_status
{
    ki_supervisord_status > /dev/null || return 1

    local SUPERVISORCTL_OUT=
    if ! SUPERVISORCTL_OUT=$(ki_supervisorctl_exec status "$@" 2>&1); then
        # Not an error if the response is STOPPED, just not running.
        if ! [[ "$SUPERVISORCTL_OUT" == *STOPPED* ]]; then
            echo_status "Error" "Executing: $KI_SUPERVISORCTL_EXE status $*"
            echo "$SUPERVISORCTL_OUT" | sed 's/^/    /g'
        fi
        return 1
    fi
}

function ki_supervisorctl_stop
{
    ki_supervisord_status > /dev/null || return 0

    local PID=
    if ! PID=$(ki_supervisorctl_pid "$@"); then
        return 0
    fi
    local ALL_PIDS=$(get_child_pids "$PID")

    local SUPERVISORCTL_OUT=
    if ! SUPERVISORCTL_OUT=$(ki_supervisorctl_exec stop "$@" 2>&1); then
        echo_status "Warning" "Executing: $KI_SUPERVISORCTL_EXE stop $*"
        echo "$SUPERVISORCTL_OUT" | sed 's/^/    /g'
    fi

    # Always check all pids, terminate if necessary.
    ALL_PIDS=$(ki_update_pids "$ALL_PIDS") || return 0
    ki_stop_pids "$ALL_PIDS" || return 1
}

function ki_supervisord_start
{
    if ki_supervisord_status > /dev/null ; then
        return 0
    fi

    ki_supervisord_stop || exit 1
    pushd "$KI_SUPERVISOR_DIR" > /dev/null
    mkdir -p "$KI_SUPERVISOR_DIR/logs" "$KI_SUPERVISOR_DIR/var"
    export HOSTNAME=$(hostname) # Used in conf files.
    export KINETICA_SUPERVISOR_ID="$KI_SUPERVISORD_CONF"
    KINETICA_PROCESS_ID="supervisord" "$KI_SUPERVISORD_EXE" -c "$KI_SUPERVISORD_CONF" -i kinetica &
    sleep 4
    ki_wait_for_cmd 60 ki_supervisord_pid 0 ki_supervisord_status 1 || true
    if ! ki_supervisord_status ; then # prints Running
        echo_status "Supervisor" "Error starting, see logs in: '$KI_SUPERVISOR_DIR/logs' for more information."
        return 1
    fi
}

function ki_supervisord_stop
{
    local ALL_PIDS=
    if ! ALL_PIDS=$(ki_supervisord_pid ALL) ; then
        \rm -f "$KI_SUPERVISORD_SOCK" || true # cleanup, supervisord logs if existing
        return 0
    fi

    SUPERVISORD_STOP_TIMEOUT=70 # long for loki
    local MAX_STOPWAITSECS=
    if MAX_STOPWAITSECS=$(grep -E "[ ]*stopwaitsecs[ ]*=[ ]*[0-9]+" "$KI_SUPERVISOR_DIR/conf/"*.conf | sed -E 's/^.*stopwaitsecs[ ]*=[ ]*([0-9]+).*$/\1/g' | sort -nr | head -n1) && \
       [ -n "$MAX_STOPWAITSECS" ]; then
        SUPERVISORD_STOP_TIMEOUT="$((MAX_STOPWAITSECS+10))"
    fi

    local PID=
    if PID=$(ki_supervisord_pid); then
        echo -n "$(echo_status "Supervisor" "Stopping ($PID) ")"
        local SHUTDOWN_MSG=
        if [ ! -e "$KI_SUPERVISORD_SOCK" ] ; then # in bad state
            echo
            echo -n "$(echo_status "Supervisor" "Terminating ")"
        elif SHUTDOWN_MSG=$(ki_supervisorctl_exec shutdown 2>&1) && \
             ki_wait_for_cmd "$SUPERVISORD_STOP_TIMEOUT" "ps -q '$ALL_PIDS'" 0; then
            echo -n " " # didn't timeout, double check all pids below
        else
            echo
            echo_status "Warning" "Supervisor had issue shutting down, terminating ($ALL_PIDS)"
            echo "$SHUTDOWN_MSG"
        fi
    else
        echo -n "$(echo_status "Supervisor" "Stopping apps ($ALL_PIDS) ")"
    fi
    sleep 4
    # Always fall though and send term to all processes remaining
    if ALL_PIDS=$(ki_update_pids "$ALL_PIDS"); then
        if ! ki_stop_pids "--progress=.,--timeout=$SUPERVISORD_STOP_TIMEOUT" "$ALL_PIDS"; then
            if ALL_PIDS=$(ki_update_pids "$ALL_PIDS"); then # Only show remaining
                echo
                echo_status "Supervisor" "Error: Unable to kill supervisor processes: $ALL_PIDS"
                exit 1
            fi
        fi
    else
        echo
    fi

    \rm -f "$KI_SUPERVISORD_SOCK" || true # supervisord logs if existing
    echo_status "Supervisor" "Stopped"
}

function ki_supervisord_stop_if_idle
{
    local ALL_PIDS=
    if ! ALL_PIDS=$(ki_supervisord_pid ALL) ; then
        return 0
    fi

    local PID=
    if PID=$(ki_supervisord_pid) && [[ "a$PID" == "a$ALL_PIDS" ]]; then
        ki_supervisord_stop || return $? # Only supervisord, no apps.
    fi
    return 0
}

function ki_supervisord_status
{
    # Check pid, client to verify socket, and 'avail' works when conf is parsed.
    local PID=
    if PID=$(ki_supervisord_pid) && \
       ki_supervisorctl_exec avail &> /dev/null; then
       echo_status "Supervisor" "Running ($PID)"
       return 0
    fi

    echo_status "Supervisor" "Stopped"
    [ -z "$PID" ] && return 1 || return 2 # 1 is stopped, 2 socket missing?
}

function ki_supervisord_pid
{
    if [[ "${1:-}" == "ALL" ]]; then
        get_pids_for_app "." "KINETICA_SUPERVISOR_ID=$KI_SUPERVISORD_CONF" || return 1
    else
        get_pids_for_app --parent "$KI_SUPERVISORD_EXE -c $KI_SUPERVISORD_CONF" "KINETICA_PROCESS_ID=supervisord" || return 1
    fi
}

# ---------------------------------------------------------------------------
# Test functions

TEST_TIME_START_SECONDS=""
TEST_NUM=""
function print_test_time
{
    echo "elapsed $((SECONDS-TEST_TIME_START_SECONDS))s, $(date '+%Y-%m-%d %H:%M:%S')"
}

function run_cmd_check_return_code
{
    local RUN_LOCAL=0
    if [ "$1" = "RUN_LOCAL" ]; then
        shift
        RUN_LOCAL=1
    fi

    local CMD="$1"
    local EXPECTED_RET_CODE="$2"
    local RET_CODE=0

    if [ -z "$TEST_TIME_START_SECONDS" ]; then
        TEST_TIME_START_SECONDS="$SECONDS"
    fi

    if [ -z "$TEST_NUM" ]; then TEST_NUM=0; fi
    TEST_NUM=$((TEST_NUM + 1))
    echo
    # Echo datetime to help find issues in the log
    echo_status "-TESTING $TEST_NUM" "$CMD : $(print_test_time)"
    local RUN_CMD_START_SECONDS="$SECONDS"

    if [ $RUN_LOCAL -eq 1 ]; then
        $CMD || RET_CODE=$?
    else
        bash -c "$CMD" || RET_CODE=$?
    fi

    if { [ "$EXPECTED_RET_CODE" -eq 0 ] && [ "$RET_CODE" -ne 0 ]; } || \
       { [ "$EXPECTED_RET_CODE" -ne 0 ] && [ "$RET_CODE" -eq 0 ]; }; then
        echo "Error running cmd: '$CMD', expected return value of '$EXPECTED_RET_CODE', got '$RET_CODE', exiting."
        exit 1
    fi

    echo_status "-SUCCESS $TEST_NUM" "in $((SECONDS-RUN_CMD_START_SECONDS))s, $(print_test_time)"
}

# ---------------------------------------------------------------------------

if [[ "${BASH_SOURCE[0]}" != "${0}" ]]; then
    #echo "${BASH_SOURCE[0]} is being sourced..."
    return
else
    "$@"
fi
